:-module(link2tree, [link2tree/3, linkage2constituents/4, chopt/1,
                     lp/3, id/3, feature/3, m/1, test/0, al2t/2]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% link2tree.pl
% converts strings via linkages into constituent trees
% main interface
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% May 2002
% Stefan Hoefler
% shoefler@cl.unizh.ch
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This version of link2tree combines with ExtrAns 1.7
% (which makes use of Link Grammar 4.1)
% For a detailled description see my master's thesis.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(linkage2links, [linkage2links/2]).
:-use_module(merge2tree, [get_tree_of_links/1]).
:-use_module(nicetree, [wordorder/2, nodelabels/2]).
:-use_module(draw, [draw/1]). 
:-use_module(linkinfo, [change_option/1, reload_linkfeatures/0]).
:-use_module(features, [feature/2]).
:-use_module('../link_grammar.pl', [parse/2, print_links/1]).

%-------------------------------------------+
% chopt(+OptionName) or chopt(+Option=Value)
%-------------------------------------------+
% changes program options stored in linkfeatures.pl

chopt(Option):-change_option(Option). % from module linkinfo

m(Model):-chopt(model=Model).

%-------------------------------------------+
% echo_option(?Option=?Value)
%-------------------------------------------+
% displays program options stored in linkfeatures.pl

echo_option(Option=Value):-if_option(Option=Value). % from module
													% linkinfo 

%-------------------------------------------------------+
% link2tree(+String, +ResultLinkageNr, -ConstituentTree)
%-------------------------------------------------------+
% simplified display of results for user queries
% ?- link2tree("cp copies the file.", Tree).

link2tree(String, ResultNr, ConstituentTree):-
	parse(String, Linkages), !,  % from main.pl
	print_links(Linkages), !,  % from main.pl
	nl, nl, write(Linkages), % test!
	linkage2constituents(0, Linkages, ResultNr, ConstituentTree), !,
	display_tree(ConstituentTree),
	!.
%---------------------------------------------------------------+
% linkage2constituents(+SentenceNr, +Linkages, +ResultLinkageNr,
% -ConstituentTree) 
%---------------------------------------------------------------+
% interface to other modules/programs
% input:  Link Grammar linkage as returned by ExtrAns 1.7
%         parse/2
% output: constituent tree of the form
%         [NodeNr1,[NodeNr2,'word2'],[NodeNr3, 'word3']]
%         as well as the features to the nodes (dynamically asserted
%         in the file features.pl: feature(NodeNr, Feature=Value).

linkage2constituents(SentNr, Linkages, ResultNr, ConstituentTree):-
	reload_linkfeatures,
	linkage2links(Linkages, ResultNr), !,  % from module linkage2links
	get_tree_of_links(UnarrangedTree), !,  % from module merge2tree
	wordorder(UnarrangedTree, ConstituentTree), !,
                                           % from module nicetree
	nl, nl, write(UnarrangedTree),          % check
	nl, nl, write(ConstituentTree), nl, nl, % check
	postprocess(SentNr, ConstituentTree).	

%-------------------------------+
% display_tree(+ConstituentTree)
%-------------------------------+

% display of the resulting constituent tree:
% - lists dynamically asserted features features:feature/2
% - draws tree using draw.pl

display_tree(ConstituentTree):-
	nodelabels(ConstituentTree, LabelledTree), % from module nicetree
	nl, nl, use_draw(LabelledTree),
	nl, nl,
	!.

%==================================================================+
% use_draw
% this predicate makes use of the module draw by Mark Holcomb.
% Draw.pl displays tree-structures in a form similar to
% [s,[np,[d,the],[n,dog]],[vp,[v,ran],[np,[d,the],[n,house]]]].
%
%==================================================================+

%--------------------+
% use_draw(+ListTree)
%--------------------+
% prints an ascii diagram of the tree

use_draw(ListTree):-
	recursive_unif(ListTree, FunctorTree),
	draw(FunctorTree). % from module draw

use_draw(_):-
	nl, nl, write("ERROR in link2tree:use_draw: Tree could not be drawn.").

%----------------------------------------+
% recursive_unif(+ListTree, -FunctorTree)
%----------------------------------------+
% changes a list tree of the form [Node, DaughterTree1, DaughterTree2]
% into a functor tree of the form node(DaughterTree1, Daughtertree2)
% using the predicat unif: Term=..[Functor|Arguments].

recursive_unif([Node|ListRest], FunctorTree):-
	list_unif(ListRest, FunctorRest),
	FunctorTree=..[Node|FunctorRest],
	!.

recursive_unif(Leaf, Leaf).

%-----------------------------------------------------+
% list_unif(+List_of_ListTrees, -List_of_FunctorTrees)
%-----------------------------------------------------+

list_unif([], []).

list_unif([FirstListTree|ListRest],[FirstFunctorTree|FunctorRest]):-
	recursive_unif(FirstListTree, FirstFunctorTree),
	list_unif(ListRest, FunctorRest).

%=========================================================+
% postprocess(+SentenceNr, +TreeInListForm)
%=========================================================+
% brings the output into the canonical form we decided on:
% linear precedence: lp(SentNr, LeftSister, RightSister)
% immediate dominance: id(SentNr, Mother, Child)
% features: feature(SentNr, NodeNr, Feature=Value)
%=========================================================+

:-dynamic(id/3).
:-dynamic(lp/3).
:-dynamic(feature/3).

postprocess(SentNr, TreeList):-
	% retracting dynamic facts from the previous query
	my_retractall(id(_,_,_)), % from module linkage2links
	my_retractall(lp(_,_,_)),
	my_retractall(feature(_,_,_)),
	% generating dynamic lp, id and feature facts
	tree2id(SentNr, TreeList),
	tree2lp(SentNr, TreeList),
	add_sentnr_to_features(SentNr),
	% printing the generated facts on the screen
	listing(id/3),
	listing(lp/3),
	listing(feature/3).

%--------------------------------------------------------+
% tree2id(+SentenceNr, +Tree)
%--------------------------------------------------------+
% generates the id/3 facts expressing immediate dominance

% case1: one or more complex child(ren)
tree2id(SentNr, [MotherID, [ChildID|Grandchildren]|RestChildren]):-
	assertz(id(SentNr, MotherID, ChildID)),
	tree2id(SentNr, [ChildID|Grandchildren]),
	tree2id(SentNr, [MotherID|RestChildren]).

% case2: leaf
tree2id(_SentNr, [_Terminal, _Leaf]).

%--------------------------------------------------------+
% tree2lp(+SentenceNr, +Tree)
%--------------------------------------------------------+
% generates the lp/3 facts expressing linear precedence

% case1: two or more children
tree2lp(S, [M, [N1|C1], [N2|C2]|R]):-
	assertz(lp(S, N1, N2)),
	tree2lp(S, [N1|C1]),
	tree2lp(S, [M, [N2|C2]|R]).

% case2: one complex child
tree2lp(S, [_M, C]):-
	tree2lp(S, C).

% case3: anything else: no child or leaf
tree2lp(_, _).

%--------------------------------------------------------+
% add_sentnr_to_features(+SentenceNr)
%--------------------------------------------------------+
% generates feature/3 facts including the sentence number
% (failure-driven loop)

add_sentnr_to_features(SentNr):-
	feature(Node, Feature=Value), % from module features
	assertz(feature(SentNr, Node, Feature=Value)),
	fail.

add_sentnr_to_features(_).

%=============================+
% general (non-iso) predicates
%=============================+

%-----------------------+
% my_retractall(+Clause)
%-----------------------+

my_retractall(Clause):-
	retract(Clause),
	fail.

my_retractall(_).