:-module(linkage2links, [test/0, linkage2links/2,
                         token/2, link/4, add_link/4, remove_link/4,
                         add_token/3, remove_token/2, add_features/2,
                         leftendpos/3, rightendpos/3, remote_head/4,
                         sub_root/2, new_nr/2]). 

%===================================================================+
% linkassert: changes the linkage-output of linkgrammar into:
% - token(Word, FeatureMatrixNr)
% - link(HeadNr, DependentNr, Linktype, ListOfSubscripts)
%===================================================================+

:-use_module(features).
:-use_module(linkinfo,[split_linktype/3, get_tagfeatures/3,
					   call_rule/0, get_treelevel/5]).
:-dynamic(link/4).
:-dynamic(token/2).
:-dynamic(last_tokennr/1).
:-dynamic(last_nr/2).

%-------------------------------------------------+
% linkage2links(+Resultlinkages, +ResultLinkageNr)
%-------------------------------------------------+

linkage2links(ResultLinkages, ResultNr):-
	retract_all_dynamics,
	get_xth_result(ResultNr, ResultLinkages, Linkage),
	get_lists(Linkage, TokenList, LinkList),
	create_list_of_links(LinkList, ListOfLinks),  
	assert_links(ListOfLinks),
	assert_tokens(TokenList),
	do_relinking,
	do_linksorting.

%---------------------+
% retract_all_dynamics
%---------------------+

retract_all_dynamics:-
	my_retractall(link(_,_,_,_)),
	my_retractall(token(_,_)),
	my_retractall(last_tokennr(_)),
	my_retractall(last_nr(_,_)),
	my_retractall(feature(_,_)),   % aus module features
	my_retractall(last_matrixnr(_)). % aus module features

%-------------------------------------------------------+
% get_xth_result(+X, +ResultLinkages, -XthResultLinkage)
%-------------------------------------------------------+

% case1: no result linkages
get_xth_result(_, [], []).

% case2: X is 0 -> get the first result linkage
get_xth_result(0, [FirstLinkage|_RestLinkages], FirstLinkage).

% case3: X > 0 -> check next result linkage and decrease X
get_xth_result(X, [_FirstLinkage|RestLinkages], XthLinkage):-
	X > 0,
	Y is X-1,
	get_xth_result(Y, RestLinkages, XthLinkage).

% case4: X < 0 -> ERROR
get_xth_result(X, _, _):-
	X < 0,
	nl,
	write('ERROR: Number of selected result linkage is negative!'),
	fail.

%-------------------------------------------+
% get_lists(+Linkage, -TokenList, -LinkList)
%-------------------------------------------+

get_lists([[TokenList, LinkList]], TokenList, LinkList).

%----------------------------------------------+
% create_list_of_links(+LinkList, -ListOfLinks)
%----------------------------------------------+

% changes the form of the LinkList from
% [[1,2,h(r),'Ss']] to [link(2,1,'S',[s])]
% in order to simplify assertion of the dynamic link/4 predicate.

create_list_of_links([], []).

create_list_of_links([FirstLink|RestLinks],
					 [FirstLinkPredicate|RestLinkPredicates]):- 
	transform_link(FirstLink, FirstLinkPredicate),
	create_list_of_links(RestLinks, RestLinkPredicates).

%--------------------------------------+
% transform_link(+Link, -Linkpredicate)
%--------------------------------------+

% changes the form of a Link from
% [1,2,h(r),'Ss'] to link(2,1,'S',[s])

% case1: Head is left
transform_link([HeadNr, DependentNr, h(l), Linktype], link(HeadNr,
														   DependentNr, Type, SubsList)):-
	split_linktype(Linktype, Type, SubsList). % aus module linkinfo

% case2: Head is right
transform_link([DependentNr, HeadNr, h(r), Linktype], link(HeadNr,
														   DependentNr, Type, SubsList)):-
	split_linktype(Linktype, Type, SubsList). % aus module linkinfo

% case3: No Head (no directionality added to this linktype)
% ATTENTION: these linktype are automatically changed into
% left-head-links! this is arbitrary!
% they have to be corrected by an appropriate rule in module linkfeatures!
transform_link([HeadNr, DependentNr, AnyOtherOrNoDirectionality,
				Linktype], link(HeadNr, DependentNr, Type,
								SubsList)):-
	split_linktype(Linktype, Type, SubsList), % aus module linkinfo
	nl, write('Linktype '), write(Linktype),
	write(' with directionality '), write(AnyOtherOrNoDirectionality),
	write(' arbitrarily changed into left-head-link!'), nl,
	write('Make shure that a correction-rule for this linktype is '),
	write('provided in linkfeatures.pl!'), nl. 
			  

%---------------------------+
% assert_links(+ListOfLinks)
%---------------------------+

% asserts all clauses link/4 of a list

assert_links([]).

assert_links([FirstLinkClause|RestLinkClauses]):-
	assertz(FirstLinkClause),
	nl, write('Link asserted: '), write(FirstLinkClause), nl,
	assert_links(RestLinkClauses).

%-----------------------------+
% assert_tokens(+TokenList)
%-----------------------------+

assert_tokens([]).

assert_tokens([Token|RestTokens]):-
	new_nr(token, TokenNr),
	add_token(Token, [position=TokenNr, cat='?'], _NewMatrixNr),
	assert_tokens(RestTokens).

%=============================+
% general (non-iso) predicates
%=============================+

%-----------------------+
% my_retractall(+Clause)
%-----------------------+

my_retractall(Clause):-
	retract(Clause),
	fail.

my_retractall(_).

%==================+
% relinking process
%==================+

%-------------+
% do_relinking
%-------------+

do_relinking:-
	call_rule, % from module linkinfo
	fail.

do_relinking.

%------------------------------+
% call_list(+ListOfExectuables)
%------------------------------+

call_list([]).

call_list([FirstTerm|RestTerms]):-
	call(FirstTerm),
	call_list(RestTerms).

%============================+
% predicates for relink-rules
%============================+
% always use cuts ! in the end of these rules!

%----------------------------------------------+
% remove_link(+HeadNr, +DepNr, +Type, +SubsList)
%----------------------------------------------+

% case1: link exists
remove_link(HeadNr, DepNr, Type, SubsList):-
	retract(link(HeadNr, DepNr, Type, SubsList)),
	nl,
	write('Link removed: '),
	write(link(HeadNr, DepNr, Type, SubsList)),
	nl,
	!.

% cas2: link does not exist
remove_link(HeadNr, DepNr, Type, SubsList):-
	nl,
	write('ERROR in linkinfo:relink: '),
	write(link(HeadNr, DepNr, Type, SubsList)),
	write(' could not be removed.'),
	nl,
	!.

%-------------------------------------------+
% add_link(+HeadNr, +DepNr, +Type, +SubsList)
%-------------------------------------------+

add_link(HeadNr, DepNr, Type, SubsList):-
	assertz(link(HeadNr, DepNr, Type, SubsList)),
	nl, write('Link asserted: '),
	write(link(HeadNr, DepNr, Type, SubsList)),
	nl,
	!.

%-----------------------------------+
% remove_token(+TokenWord, +MatrixNr)
%-----------------------------------+

% case1: token exists
remove_token(Token, Matrix):-
	retract(token(Token, Matrix)),
	my_retractall(feature(Matrix,_)),
	nl, write('Token removed: '),
	write(token(Token, Matrix)),
	nl,
	!.

% case2: token does not exist
remove_token(Token, Matrix):-
	nl,
	write('ERROR in linkinfo:relink: '),
	write(token(Token, Matrix)),
	write(' could not be removed.'),
	nl,
	!.

%-----------------------------------------------+
% add_token(+TokenWord, +FeatureList, -MatrixNr)
%-----------------------------------------------+

add_token(Token, FeatureList, MatrixNr):-
	create_matrix(FeatureList, MatrixNr), % from module features
	get_tagfeatures(Token, TagFeatureList, Word), % from module linkinfo
	add_featurelist_to_matrix(MatrixNr, [hd=Word|TagFeatureList]),
                                          % from module features
	assertz(token(Word, MatrixNr)),
	nl, write('Token asserted: '),
	write(token(Word, MatrixNr)),
	nl,
	!.

%-----------------------------------+
% add_features(+Matrix, +FeatureList)
%-----------------------------------+

add_features(Matrix, FeatureList):-
	add_featurelist_to_matrix(Matrix, FeatureList), % from module
                                                    % features
	!.

%----------------------------------------------------+
% remote_head(+Dep, +TopType, +TopSubsList, -TopHead)
%----------------------------------------------------+
%
% finds TopHead-(TopType,TopSubs)->Dep1->Dep2->...->Dep

% case1: TopHead is direct head
remote_head(Dep, TopType, TopSubs, TopHead):-
	link(TopHead, Dep, TopType, TopSubs),
	!.

% case2: recursive
remote_head(Dep, TopType, TopSubs, TopHead):-
	link(Head, Dep, _, _),
	remote_head(Head, TopType, TopSubs, TopHead),
	!.

% case3: no TopHead of TopType/TopSubs exists
remote_head(_Dep, TopType, TopSubs, _):-
	nl,
	write('ERROR in relinking rule: No TopHead of Type: '),
	write(TopType), write(TopSubs),
	write(' exists. Check rule in module linkfeatures.'),
	nl,!.

%-----------------------------------+
% sub_root(+Dependent, -SubRootWord)
%-----------------------------------+
%
% finds the root word of a linkage starting from one token

% case1: there is a head
sub_root(Token, RootWord):-
	link(Head, Token, _, _),
	sub_root(Head, RootWord),
	!.

% case2: no head left -> token is root word
sub_root(RootWord, RootWord):-!.

%------------------------+
% new_nr(+NrType, -NewNr)
%------------------------+

% case1: last nr of this type existing
new_nr(NrType, NewNr):-
	last_nr(NrType, LastNr),
	NewNr is LastNr +1,
	retract(last_nr(NrType, LastNr)),
	assertz(last_nr(NrType, NewNr)),
	!.

% case2: initialise nr of this type
new_nr(NrType, InitNr):-
	InitNr is 0, % initialise nr
	assertz(last_nr(NrType, InitNr)),
	!.

%=============================================+
% leftendpos(+HeadPos, +ProjType, -LeftEndPos) 
%=============================================+

leftendpos(HeadPos, ProjType, LeftEndPos):-
	mindeppos(HeadPos, ProjType, MinDepPos),
	prevpos(MinDepPos, PrevPos),
	LeftEndPos is (MinDepPos+PrevPos)/2,
	!.

%-------------------------------------------+
% mindeppos(+HeadPos, +ProjType, -MinDepPos)
%-------------------------------------------+

% case1: link of this ProjType exists
mindeppos(HeadPos, ProjType, MinDepPos):-
	id2pos(HeadID, HeadPos),
	link(HeadID, DepID, _, ProjType),
	id2pos(DepID, DepPos),
	DepPos<HeadPos,
	smallestdeppos(DepPos, HeadPos, ProjType, MinDepPos),
	!.

% case2: no link of this ProjType exists -> next ProjType 
mindeppos(HeadPos, ProjType, MinDepPos):-
	get_treelevel(ProjType, _, _, NextProjType, _),
	mindeppos(HeadPos, NextProjType, MinDepPos),
	!.

% case3: this ProjType is not defined
mindeppos(HeadPos, _ProjType, HeadPos).

%---------------------------------------------------------------+
% smallestdeppos(+StartPos, +HeadPos, +ProjType, -SmallestDepPos)
%---------------------------------------------------------------+

% case1: a link of this ProjType exists
smallestdeppos(StartPos, HeadPos, ProjType, SmallestDepPos):-
	id2pos(HeadID, HeadPos),
	link(HeadID, DepID, _, ProjType),
	id2pos(DepID, DepPos),
	StartPos>DepPos,
	smallestdeppos(DepPos, HeadPos, ProjType, SmallestDepPos),
	!.

% case2: no link of this ProjType with smaller DepPos can be found
smallestdeppos(StartPos, _, _, StartPos).

%------------------------+
% prevpos(+Pos, -PrevPos)
%------------------------+

% case1: there is a previous position
prevpos(X, Y):-
	feature(_, position=Y),
	X>Y,
	\+intermediate(X, Y),
	!.

% case2: there is no previous Pos
prevpos(X, Y):-
	Y is X-1.

%===============================================+
% rightendpos(+HeadPos, +ProjType, -RightEndPos) 
%===============================================+

rightendpos(HeadPos, ProjType, RightEndPos):-
	maxdeppos(HeadPos, ProjType, MaxDepPos),
	nextpos(MaxDepPos, NextPos),
	RightEndPos is (MaxDepPos+NextPos)/2,
	!.

%-------------------------------------------+
% maxdeppos(+HeadPos, +ProjType, -MaxDepPos)
%-------------------------------------------+

% case1: link of this ProjType exists
maxdeppos(HeadPos, ProjType, MaxDepPos):-
	id2pos(HeadID, HeadPos),
	link(HeadID, DepID, _, ProjType),
	id2pos(DepID, DepPos),
	DepPos>HeadPos,
	biggestdeppos(DepPos, HeadPos, ProjType, MaxDepPos),
	!.

% case2: no link of this ProjType exists -> next ProjType 
maxdeppos(HeadPos, ProjType, MaxDepPos):-
	get_treelevel(ProjType, _, _, NextProjType, _),
	maxdeppos(HeadPos, NextProjType, MaxDepPos),
	!.

% case3: this ProjType is not defined
maxdeppos(HeadPos, _ProjType, HeadPos).

%--------------------------------------------------------------+
% biggestdeppos(+StartPos, +HeadPos, +ProjType, -BiggestDepPos)
%--------------------------------------------------------------+

% case1: a link of this ProjType exists
biggestdeppos(StartPos, HeadPos, ProjType, BiggestDepPos):-
	id2pos(HeadID, HeadPos),
	link(HeadID, DepID, _, ProjType),
	id2pos(DepID, DepPos),
	StartPos<DepPos,
	biggestdeppos(DepPos, HeadPos, ProjType, BiggestDepPos),
	!.

% case2: no link of this ProjType with bigger DepPos can be found
biggestdeppos(StartPos, _, _, StartPos).

%------------------------+
% nextpos(+Pos, -NextPos)
%------------------------+

% case1: there is a next position
nextpos(X, Y):-
	feature(_, position=Y),
	X<Y,
	\+intermediate(X, Y),
	!.

% case2: there is no bigger Pos
nextpos(X, Y):-
	Y is X+1.

%================================================================+
% auxiliary predicates for the calculation of new token positions
%================================================================+

%------------------+
% id2pos(?ID, ?Pos)
%------------------+
id2pos(ID, Pos):-
	feature(ID, position=Pos).

%---------------------+
% intermediate(+X, +Y)
%---------------------+

intermediate(X, Y):-
	feature(_, position=Z),
	X<Z, Z<Y.

intermediate(X, Y):-
	feature(_, position=Z),
	X>Z, Z>Y.

%================================================================%
% do_linksorting/0
%================================================================%
% re-sorts the link/4 facts after relinking according to the order
% given by Link Grammar

do_linksorting:-
	links2list(UnsortedList), !,
	linksort(UnsortedList, SortedList), !,
	list2links(SortedList).

%-------------------------+
% links2list(-ListOfLinks)
%-------------------------+
% retracts the existing link/4 facts and reads them into a list

links2list([[H, D, T, S]|R]):-
	retract(link(H, D, T, S)),
	links2list(R).
links2list([]).

%-------------------------------------+
% linksort(+UnsortedList, -SortedList)
%-------------------------------------+
% quicksort for links

linksort([],[]).
linksort([X|Tail], Sorted):-
	linksplit(X, Tail, Small, Big),
	linksort(Small, SortedSmall),
	linksort(Big, SortedBig),
	append(SortedSmall, [X|SortedBig], Sorted).

%-----------------------------------+
% linksplit(+X, +Tail, -Small, -Big)
%-----------------------------------+
% split for links

linksplit(_X, [], [], []).
linksplit(X, [Y|Tail], [Y|Small], Big):-
	X = [HX, DX|_],
	Y = [HY, DY|_],
	get_left(HX, DX, LX, _RX),
	get_left(HY, DY, LY, _RY),
	LX > LY,
	linksplit(X, Tail, Small, Big).
linksplit(X, [Y|Tail], Small, [Y|Big]):-
	X = [HX, DX|_],
	Y = [HY, DY|_],
	get_left(HX, DX, LX, _RX),
	get_left(HY, DY, LY, _RY),
	LX < LY,
	linksplit(X, Tail, Small, Big).
linksplit(X, [Y|Tail], [Y|Small], Big):-
	X = [HX, DX|_],
	Y = [HY, DY|_],
	get_left(HX, DX, LX, RX),
	get_left(HY, DY, LY, RY),
	LX = LY,
	RX < RY,
	linksplit(X, Tail, Small, Big).
linksplit(X, [Y|Tail], Small, [Y|Big]):-
	X = [HX, DX|_],
	Y = [HY, DY|_],
	get_left(HX, DX, LX, _RX),
	get_left(HY, DY, LY, _RY),
	LX = LY,
	linksplit(X, Tail, Small, Big).

%-------------------------------------------------------+
% get_left(+Nr1, +Nr2, -SmallerPosition, -BiggerPosition)
%-------------------------------------------------------+

get_left(A, B, L, R):-
	feature(A, position=APos),
	feature(B, position=BPos),
	APos < BPos,
	L = APos,
	R = BPos, !.
get_left(A, B, L, R):-
	feature(A, position=APos),
	feature(B, position=BPos),
	L = BPos,
	R = APos, !.

%-------------------------+
% list2links(+ListOfLinks)
%-------------------------+
% asserts link/4 facts from a list of links

list2links([[H, D, T, S]|R]):-
	assertz(link(H, D, T, S)),
	list2links(R).
list2links([]).

%----------+
% append/3
%----------+

append([], List, List).
append([Element], List, [Element|List]).
append([Element|Rest], List, [Element|AppendList]):-
	append(Rest, List, AppendList).
