:-module(linkfeatures_gb, [rootlink/2, typefeatures/4, subsfeatures/4,
						projectable/1, footfeature/1, relink/0,
						tagfeatures/2, treelevel/5,
						toplevel/1, headfeature/1]). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% linkfeatures.pl
% the "grammar" module: here, all the information how links shall be
% converted into constituents is stored.
% adaptions need only be made in this module.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mai 2001
% Stefan Hoefler
% shoefler@cl.unizh.ch
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(linkage2links, [link/4, token/2, add_link/4,
							 remove_link/4, add_token/3,
							 remove_token/2, add_features/2,
							 rightendpos/3, leftendpos/3,
							 remote_head/4, sub_root/2, new_nr/2]).
:-use_module(features, [feature/2]).

%---------------------------------------------------------------------+
% treelevel(TypeOfProjection, Recursive(nil/stack/flat), obligatory/optional,
% Next(Deeper)TypeOfProjection, AdditionalFeaturesListOfTheProjection)
%---------------------------------------------------------------------+

treelevel(specifier, nil, obligatory, adjunct, [bar=2]).
treelevel(adjunct, stack, optional, complement, [bar=1]).
treelevel(complement, flat, obligatory, compound, [bar=1]).
treelevel(compound, stack, optional, [], [bar=0]).


%------------------------------+
% toplevel(TopTypeOfProjection)
%------------------------------+

toplevel(specifier).

%-----------------------------------------------+
% rootlink(RootlinkType, RootlinkSubscriptsList)
%-----------------------------------------------+

rootlink('S', _AnySubscripts).

%----------------------+
% projectable(Category)
%----------------------+

projectable(a).   % adjective phrases
projectable(adv). % adverbial phrases
projectable(c).   % complementizer phrases
projectable(d).   % determiner phrases
projectable(i).   % inflectional phrases
projectable(n).   % noun phrases
projectable(p).   % prepositional phrases
projectable(v).   % verb phrases

%------------------------------------------------+
% typefeatures(LinktypeCapitals, HeadFeatureList,
% DependentFeatureList, ProjectionType)
%------------------------------------------------+

% linktypes defined by Link Grammar 4.1
typefeatures('A', [cat=n, bar=0], [cat=a, bar=0], adjunct).
typefeatures('AF', [cat=v, bar=0], [cat=a, bar=0], complement).
typefeatures('AL',[cat=d, bar=0],[cat=d, bar=0], complement).
typefeatures('AN', [cat=n, bar=0], [cat=n, bar=0], compound).
typefeatures('AZ', [cat=v, bar=0], [cat=p, bar=0], adjunct). % double object!
typefeatures('E', [cat=v, bar=0], [], adjunct).
typefeatures('EA', [cat=a, bar=0], [cat=adv, bar=0], adjunct).
typefeatures('EE', [cat=adv, bar=0], [cat=adv, bar=0], adjunct).
typefeatures('S', [cat=i, bar=0], [cat=d, bar=0], specifier).
typefeatures('O', [cat=v, bar=0], [cat=d, bar=0], complement).
typefeatures('P', [bar=0], [bar=0], complement). % feature 'be' for the head?

% linkypes defined in this grammar
typefeatures('C', [cat=c, bar=0], [cat=i, bar=0], complement).
typefeatures('D', [cat=d, bar=0], [cat=n, bar=0], complement).
typefeatures('INFL',[cat=i, bar=0], [cat=v, bar=0],
			 complement).
typefeatures('R', [], [cat=c, bar=0], complement).
typefeatures('RS', [cat=i, bar=0], [cat=d, bar=0], specifier).
typefeatures('RWHO', [cat=c, bar=0], [cat=d, bar=0, rel=1], specifier).

%------------------------------------------------------+
% subsfeatures(+Subscript, +TypeList, -HeadFeatureList,
% -DependendtFeatureList)
%------------------------------------------------------+

subsfeatures('p', ['O'], [], [num=pl]).
subsfeatures('p', [_AnyOtherType], [num=pl], [num=pl]).
subsfeatures('s', ['O'], [], [num=sg]).
subsfeatures('s', [_AnyOtherType], [num=sg], [num=sg]).

%---------------------+
% headfeature(Feature)
%---------------------+

headfeature(bar).
headfeature(cat).
headfeature(hd).
headfeature(num).
headfeature(rel).

% do not delete this declarations!
headfeature(position). % do not remove this line!
headfeature(trace).    % do not remove this line!


%---------------------+
% footfeature(Feature)
%---------------------+

footfeature(wh).

%------------------------------------------------+
% tagfeatures(TagCharactersList, TagFeautureList)
%------------------------------------------------+
%
% attention: for affixes that begin with the same letter: write clause
% for the longer affix first
%
% source: extrans_1.7/link-4.1/data21/README_NEWTAGS_WORDS2.1

tagfeatures(['a'|_], [cat=a, bar=0]).
tagfeatures(['e'|_], [cat=adv, bar=0]).
tagfeatures(['n'|_], [cat=n, bar=0]).
tagfeatures(['o'|_], [cat=p, bar=0]).
tagfeatures(['p'|_], [cat=d, bar=0]).
tagfeatures(['v'|_], [cat=v, bar=0]).
tagfeatures(_,[]). % default: do not delete!

%----------------------------+
% relink
%----------------------------+
%
% ATTENTION: order of relink-rules may be significant

spec_head_pos(SpecPos, HeadPos):-
	number(HeadPos),
	leftendpos(HeadPos, complement, SpecPos).
spec_head_pos(SpecPos, HeadPos):-
	number(SpecPos),
	rightendpos(SpecPos, specifier, HeadPos).

head_compl_pos(HeadPos, ComplPos):-
	number(HeadPos),
	rightendpos(HeadPos, complement, ComplPos).
head_compl_pos(HeadPos, ComplPos):-
	number(ComplPos),
	leftendpos(ComplPos, specifier, HeadPos).

relink:-
	link(HeadNr, DepNr, 'W', Subs),
	remove_link(HeadNr, DepNr, 'W', Subs).

relink:-
	link(HeadNr, DepNr, 'RW', Subs),
	remove_link(HeadNr, DepNr, 'RW', Subs).

relink:-
	link(HeadNr, DepNr, 'X', Subs),
	remove_link(HeadNr, DepNr, 'X', Subs).

relink:-
	link(HeadNr, DepNr, 'AA', Subs),
	remove_link(HeadNr, DepNr, 'AA', Subs).
 

% insert empty inflection to build IP
relink:-
	link(VerbNr, SubjNr, 'S', Subs),
	\+link(VerbNr, _, 'P', _),
	\+link(VerbNr, _, 'INFL', _),
	remove_link(VerbNr, SubjNr, 'S', Subs),
	feature(VerbNr, position=VerbPos),
	head_compl_pos(InflPos, VerbPos),
	add_token([],[position=InflPos, cat=i], InflNr),
	add_link(InflNr, SubjNr, 'S', Subs),
	add_link(InflNr, VerbNr, 'INFL', Subs).

% insert empty inflection to build IP:
% for RS-links in relativ clauses
relink:-
	link(WhoNr, VerbNr, 'RS', Subs),
	\+link(VerbNr, _, 'P', _),
	\+link(VerbNr, _, 'INFL', _),
	remove_link(WhoNr, VerbNr, 'RS', Subs),
	feature(VerbNr, position=VerbPos),
	head_compl_pos(InflPos, VerbPos),
	add_token([],[position=InflPos, cat=i], InflNr),
	add_link(InflNr, WhoNr, 'RS', Subs),
	add_link(InflNr, VerbNr, 'INFL', Subs).

% apply DP-hypthesis (invert D-links)
relink:-
	link(NounNr, DetNr, 'D', DSubs),
	remove_link(NounNr, DetNr, 'D', DSubs),
	remove_link(HeadNr, NounNr, HeadType, HeadSubs),
	add_link(DetNr, NounNr, 'D', DSubs),
	add_link(HeadNr, DetNr, HeadType, HeadSubs).

% apply DP-hypthesis (insert empty det to noun)
relink:-
	token(_Noun, NounNr),
	feature(NounNr, cat=n),
	\+link(NounNr, _, 'D', _),
	\+link(_, NounNr, 'D', _),
	\+link(_, NounNr, 'AN', _),
	remove_link(HeadNr, NounNr, HeadType, HeadSubs),
	feature(NounNr, position=NounPos),
	head_compl_pos(DetPos, NounPos),
	add_token([],[position=DetPos, cat=d],DetNr),
	add_link(HeadNr, DetNr, HeadType, HeadSubs),
	add_link(DetNr, NounNr, 'D', HeadSubs).

%..............................%
% Relative Clauses (R, Cr, RS) %
%..............................%

% relative clauses (R, Cr) with that in object position:
% I saw the dog that I chased.
relink:-
	link(Noun, That, 'R', RelSubs),
	link(That, RelSubj, 'C', [r|_]),
	link(RelInfl, RelSubj, 'S', SubSubs),
	token('that.c', That),
	% R-link is correct; rename it
	remove_link(Noun, That, 'R', RelSubs),
	add_link(Noun, That, 'R', RelSubs),
	% relink C-link to RelInfl
	remove_link(That, RelSubj, 'C', [r|_]),
	add_link(That, RelInfl, 'C', []),
	% rename S-link to RS-link (must not be rootlink!)
	remove_link(RelInfl, RelSubj,  'S', SubSubs),
	add_link(RelInfl, RelSubj, 'RS', SubSubs),
	% add traces
	new_nr(trace, Trace),
	link(Det, Noun, 'D', _),
	add_features(Det, [trace=Trace]),
	link(Noun, RelVerb, 'B', _),
	feature(RelVerb, position=RelVerbPos),
	head_compl_pos(RelVerbPos, ObjectPos),
	add_token([], [position=ObjectPos, trace=Trace], Object),
	add_link(RelVerb, Object, 'O', []).
	

% relative clauses (R, Cr) with who/whom in object position:
% I saw the dog who I chased.
relink:-
	link(Noun, Who, 'R', RelSubs),
	link(Who, RelSubj, 'C', [r|_]),
	link(RelInfl, RelSubj, 'S', SubSubs),
	\+token('that.c', Who),
	% insert empty C
	feature(Who, position=WhoPos),
	spec_head_pos(WhoPos, CPos),
	add_token([], [position=CPos, cat=c, bar=0], C),
	% relink R to C
	remove_link(Noun, Who, 'R', RelSubs),
	add_link(Noun, C, 'R', RelSubs),
	% relink C
	remove_link(Who, RelSubj, 'C', [r|RestCSubs]),
	add_link(C, RelInfl, 'C', [r|RestCSubs]),
	% new link from C to Who
	add_link(C, Who, 'RWHO', RelSubs),
	% change S-link into RS-link
	remove_link(RelInfl, RelSubj,  'S', SubSubs),
	add_link(RelInfl, RelSubj, 'RS', SubSubs).

% relative clauses (R, RS) with that in subject position:
% I saw the dog that barked.
relink:-
	link(Noun, That, 'R', RelSubs),
	link(RelInfl, That, 'RS', SubSubs),
	token('that.c', That),
	% insert empty relative subject
	feature(RelInfl, position=RelInflPos),
	spec_head_pos(RelSubjPos, RelInflPos),
	add_token([], [position=RelSubjPos, cat=d, bar=0], RelSubj),
	% R is correctly linked: change name to R
	remove_link(Noun, That, 'R', RelSubs),
	add_link(Noun, That, 'R', RelSubs),
	% add C-link from That to RelInfl
	add_link(That, RelInfl, 'C', []),
	% add subject-link from RelInfl to RelSubj; remove RS-link
	remove_link(RelInfl, Who, 'RS', SubSubs),
	add_link(RelInfl, RelSubj, 'RS', SubSubs).
	

% relative clauses (R, RS) with who/whom in subject position:
% I saw the dog who barked.
relink:-
	link(Noun, Who, 'R', RelSubs),
	link(RelInfl, Who, 'RS', SubSubs),
	\+token('that.c', Who),
	% insert empty C
	feature(Who, position=WhoPos),
	spec_head_pos(WhoPos, CPos),
	add_token([], [position=CPos, cat=c, bar=0], C),
	% insert empty relative subject
	feature(RelInfl, position=RelInflPos),
	spec_head_pos(RelSubjPos, RelInflPos),
	add_token([], [position=RelSubjPos, cat=d, bar=0], RelSubj),
	% change dep of R from Who to C
	remove_link(Noun, Who, 'R', RelSubs),
	add_link(Noun, C, 'R', RelSubs),
	% add C-link from C to RelInfl
	add_link(C, RelInfl, 'C', []),
	% add subject-link from RelInfl to RelSubj; remove RS-link
	remove_link(RelInfl, Who, 'RS', SubSubs),
	add_link(RelInfl, RelSubj, 'RS', SubSubs),
	% new link from C to Who
	add_link(C, Who, 'RWHO', RelSubs).

% relative clauses (R) without relative pronoun
% The dog I saw barked
relink:-
	link(Noun, RelSubj, 'R', RSubs),
	link(Noun, _, 'B', _), % compulsory; see comment on R link
	link(RelInfl, RelSubj, 'S', RSSubs),
	% insert empty C
	feature(Noun, position=NounPos),
	head_compl_pos(NounPos, CPos),
	add_token([], [position=CPos, cat=c, bar=0], C),
	% relink R
	remove_link(Noun, RelSubj, 'R', RSubs),
	add_link(Noun, C, 'R', RelSubs),
	% add C-link from C to RelInfl
	add_link(C, RelInfl, 'C', [r]),
	% change S into RS
	remove_link(RelInfl, RelSubj, 'S', RSSubs),
	add_link(RelInfl, RelSubj, 'RS', RSSubs),
	% add traces
	new_nr(trace, TraceNr),
	link(Det, Noun, 'D', _),
	add_features(Det, [trace=TraceNr]),
	link(Noun, RelVerb, 'B', _),
	feature(RelVerb, position=RelVerbPos),
	head_compl_pos(RelVerbPos, RelObjPos),
	add_token([], [position=RelObjPos, trace=TraceNr], RelObj),
	add_link(RelVerb, RelObj, 'O', []).
