:-module(linkinfo, [linkfeatures/5, is_rootlink/2, is_projectable/1,
                    is_footfeature/1, split_linktype/3, call_rule/0,
                    get_tagfeatures/3, change_option/1, if_option/1,
                    get_treelevel/5, get_toplevel/1,
                    is_headfeature/1, reload_linkfeatures/0]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% linkinfo.pl
% interface between the "grammar" (module linkfeatures) and the
% converter.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% April 2002
% Stefan Hoefler
% shoefler@cl.unizh.ch
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(linkfeatures).

:-dynamic(linkfeatures/4).

%-------------------------------------------------+
% linkfeatures(+Type, +SubsList, -HeadFeatureList,
% -DependentFeatureList, -ProjectionType)
%-------------------------------------------------+

linkfeatures(Type, SubsList, HeadFeatureList, DependentFeatureList,
			 ProjectionType):-
	get_typefeatures(Type, HeadFeatureList_1, DepFeatureList_1,
					 ProjectionType),
	add_subsfeatures(SubsList, Type, HeadFeatureList_1,
					 DepFeatureList_1, HeadFeatureList,
					 DependentFeatureList),
	!.

%-----------------------------------------------------------------+
% get_typefeatures(+Type, -HeadFeatureList, -DependentFeatureList,
% -ProjectionType)
%-----------------------------------------------------------------+

get_typefeatures(Type, HeadFeatureList, DepFeatureList,
				 ProjectionType):-
	typefeatures(Type, HeadFeatureList, DepFeatureList,
				 ProjectionType).

%-------------------------------------------------+
% split_linktype(+Linktype, -Type, -SubscriptList)
%-------------------------------------------------+

split_linktype(Linktype, Type, SubsList):-
	name(Linktype, CharList),
	get_capitals(CharList, CapitalsList, MinorsList),
	name(Type, CapitalsList),
	name_subscripts(MinorsList, SubsList).

%---------------------------------------------------+
% name_subscripts(+MinorsASCIIList, -SubscriptsList)
%---------------------------------------------------+

name_subscripts([], []).

name_subscripts([FirstASCII|RestASCII],[FirstSub|RestSubs]):-
	atom_chars(FirstSub, [FirstASCII]),
	name_subscripts(RestASCII, RestSubs).

%------------------------------------------------------+
% get_capitals(+CharList, -CapitalsList, -RestCharList)
%------------------------------------------------------+

% case1: empty character list
get_capitals([], [], []).  % capitals aus einer leeren Liste

% case2: first element is a capital
get_capitals([FirstChar|RestChars], [FirstChar|RestCapitals],
			 RestCharList):-
	65 =< FirstChar, FirstChar =< 90,
	get_capitals(RestChars, RestCapitals, RestCharList),
	!.

% case3: first element is $
get_capitals([FirstChar|RestChars], [FirstChar|RestCapitals],
			 RestCharList):-
	36=:=FirstChar,
	get_capitals(RestChars, RestCapitals, RestCharList),
	!.

% case4: first element is no capital (nor $)
get_capitals(RestChars, [], RestChars).  % subscripts

%-----------------------------------------------------------+
% add_subsfeatures(+SubscriptsList, +Type, +HeadFeatureList,
% +DependentFeatureList, -CompleteHeadFeautureList,
% -CompleteDependentFeatureList)
%-----------------------------------------------------------+

% if no subscritps left
add_subsfeatures([], _Type, HeadFeatureList, DepFeatureList,
                 HeadFeatureList, DepFeatureList).

% there are subscripts left
add_subsfeatures([Subscript|RestSubs], Type, HeadFeatureList_1,
                 DepFeatureList_1, HeadFeatureList, DepFeatureList):-
	get_subsfeatures(Subscript, Type, SubsHeadFeatureList, SubsDepFeatureList),
	append_featurelist(HeadFeatureList_1, SubsHeadFeatureList,
					   HeadFeatureList_2),
	append_featurelist(DepFeatureList_1, SubsDepFeatureList,
					   DepFeatureList_2),
	add_subsfeatures(RestSubs, Type, HeadFeatureList_2, DepFeatureList_2,
					 HeadFeatureList, DepFeatureList).

%----------------------------------------------------------+
% get_subsfeatures(+Subscript, +Type, -SubsHeadFeatureList,
% -SubsDepFeatureList)
%----------------------------------------------------------+

get_subsfeatures(Subscript, Type, SubsHeadFeatureList,
				 SubsDepFeatureList):-
	subsfeatures(Subscript, TypeList, SubsHeadFeatureList,
				 SubsDepFeatureList),
	element_of_list(Type, TypeList).

get_subsfeatures(_Subscript, _Type, [], []).  % unknown subscripts

%---------------------------------+
% element_of_list(+Element, +List)
%---------------------------------+

element_of_list(Element, [Element|_Rest]).
element_of_list(Element, [_FirstElement|Rest]):-
	element_of_list(Element, Rest).

%--------------------------------------------------+
% append_featurelist(+ListA, +ListB, -CombinedList)
%--------------------------------------------------+

append_featurelist([], List, List).
append_featurelist([Element], List, [Element|List]).
append_featurelist([First|Rest], List, [First|AppendList]):-
	append_featurelist(Rest, List, AppendList).

%---------------------+
% is_projectable(?Cat)
%---------------------+

is_projectable(Cat):-
	projectable(Cat).   % aus module linkfeatures

%------------------------------------------------+
% is_rootlink(+RootlinkType, +RootSubscriptsLink)
%------------------------------------------------+

is_rootlink(RootType, RootSubsList):- % wenn ein Rootlink vorhanden
	rootlink(RootType, RootSubsList). % aus module linkfeatures

%-------------------------+
% is_footfeature(?Feature)
%-------------------------+

is_footfeature(Feature):-
	footfeature(Feature).   % aus module linkfeatures

%-------------------------+
% is_headfeature(?Feature)
%-------------------------+

is_headfeature(Feature):-
	headfeature(Feature).   % aus module linkfeatures

%-----------------------------+
% call_rule
%-----------------------------+
% interface to module linkfeatures

call_rule:-
	relink.	% from module linkfeatures

%-------------------------+
% get_treelevel(_,_,_,_,_)
%-------------------------+
% interface to module linkfeatures

get_treelevel(ProjType, Recursive, Obligatory, NextProjType, ProjFeatures):-
	treelevel(ProjType, Recursive, Obligatory, NextProjType, ProjFeatures).

%--------------------------+
% get_toplevel(TopProjType)
%--------------------------+

get_toplevel(TopProjType):-
	toplevel(TopProjType).

%=================+
% handling options
%=================+

%---------------------------------------------------------------+
% change_option(+OptionName) or change_option(+OptionName=Value)
%---------------------------------------------------------------+

% case1: Option=Value, Option already exists
change_option(Option=NewValue):-
	option(Option=OldValue), % from module linkfeatures
	retract(option(Option=OldValue)),
	assertz(option(Option=NewValue)),
	nl, write('Option '), write(Option), write(' set to '),
	write(NewValue), nl.

% case2: only Option indicated, Option set to 0
change_option(Option):-
	option(Option=0), % from module linkfeatures
	retract(option(Option=0)),
	assertz(option(Option=1)),
	nl, write('Option '), write(Option), write(' set to 1'), nl.

% case3: only Option indicated, Option set to 1
change_option(Option):-
	option(Option=1), % from module linkfeatures
	retract(option(Option=1)),
	assertz(option(Option=0)),
	nl, write('Option '), write(Option), write(' set to 0'), nl.

% case4: error in option handling
change_option(_):-
	nl,
	write('ERROR: Option does not exist or is not used in this way.'),
	nl, write('Check module linkfeatures.'), nl.

%---------------------------+
% if_option(+Option=Value)
%---------------------------+

if_option(Option=Value):-
	option(Option=Value). % from module linkfeatures

%==================================================+
% get the features of tokens out of its affixed tag
%==================================================+

%-------------------------------------------------+
% get_tagfeatures(+Token, -TagFeaturesList, -Word)
%-------------------------------------------------+

get_tagfeatures(Token, TagFeatures, Word):-
	atom_chars(Token, CharList),
	check4features(CharList, TagFeatures, Word),
	!.

%-----------------------------------------------------+
% check4features(+ASCIIList, -TagFeauturesList, -Word)
%-----------------------------------------------------+

check4features(TokenASCIIList, TagFeatures, Word):-
	get_tag(TokenASCIIList, TagASCIIList, WordASCIIList),
	make_chars_list(TagASCIIList, TagCharsList),
	atom_chars(Word, WordASCIIList),
	tagfeatures(TagCharsList, TagFeatures).	% from module linkfeatures

check4features(_, []).

%------------------------------------------------------+
% get_tag(+WordASCIIList, -TagASCIIList, -WordformList)
%------------------------------------------------------+

% case1: empty ASCII-List
get_tag([],[], []).

% case2a: a full stop is the only symbol
get_tag([FullStop], [], [FullStop]):-
	FullStop=:=46.

% case2b: first symbol is a dot -> rest is tag
get_tag([FirstASCII|Tag],Tag, []):-
	FirstASCII=:=46.

% case3: cut first symbol
get_tag([FirstASCII|RestASCII],Tag, [FirstASCII|RestWordform]):- 
	get_tag(RestASCII, Tag, RestWordform).
	

%----------------------------------------+
% make_chars_list(+ASCIIList, -CharsList)
%----------------------------------------+

make_chars_list([],[]).

make_chars_list([ASCII|RestASCII],[Char|RestChars]):-
	atom_chars(Char, [ASCII]),
	make_chars_list(RestASCII, RestChars).

%==================================================================+
% reload_linkfeatures/0: load a another rule set during the program
%==================================================================+

% case1: the option model has not been changed since the last time
reload_linkfeatures:-
	option(last_model=Model),
	option(model=Model).

% case2: the option model has been changed since the last time
reload_linkfeatures:-
	abolish([rootlink/2, typefeatures/4, subsfeatures/4,
             projectable/1, footfeature/1, relink/0, tagfeatures/2,
             treelevel/5, toplevel/1, headfeature/1]),
	option(model=Model),
	modelpath(Model, ModelPath),
	use_module(ModelPath, [rootlink/2, typefeatures/4, subsfeatures/4,
                          projectable/1, footfeature/1, relink/0,
                          tagfeatures/2, treelevel/5, toplevel/1,
                          headfeature/1]),
	change_option(last_model=Model).

