:-module(merge2tree, [get_tree_of_links/1]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% merge2tree.pl - converts links into a tree
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(linkinfo, [get_toplevel/1, get_treelevel/5,
                        is_projectable/1, is_rootlink/2,
                        is_footfeature/1, linkfeatures/5,
                        is_headfeature/1]). 
:-use_module(linkage2links, [new_nr/2, link/4, token/2]).
:-use_module(features).

%-------------------------+
% get_tree_of_links(-Tree)
%-------------------------+
% gets the root link type and checks if a link of this type
% exists. The head of this link is the starting point of the
% conversion process.

get_tree_of_links(Tree):-
	get_rootlink(RootType, RootSubsList),
	link(StartHeadNr, _DependentNr, RootType, RootSubsList),
                                      % from module linkage2links
	merge2tree(StartHeadNr, Tree).

%---------------------------+
% merge2tree(+HeadNr, -Tree)
%---------------------------+

% case1: the category of the head is projectable
% - gets the top projection type (eg. specifier) (get_toplevel/1)
% - resolves (retracts) all links from this head of the projection type "trace" and
% adds a trace index to their tokens (trace/1).
% - starts the projection of a tree with HeadNr as its head


merge2tree(HeadNr, Tree):-
	cat_is_projectable(HeadNr),
	get_toplevel(ProjType), % from module linkinfo 
	trace(HeadNr),
	projection(ProjType, HeadNr, Tree),
	!.

% case2: the category of the head is not projectable
% - turn the head into a terminal subtree

merge2tree(HeadNr, LeafTree):-
	add_leaf(HeadNr, LeafTree).

%---------------+
% trace(+HeadNr)
%---------------+
% resolves (retracts) all links from this head of the projection type "trace" and
% initialises and adds trace indices to the feature structure of their
% tokens (trace/1).

% case1: if trace exists
trace(HeadNr):-
	link(HeadNr, TraceNr, Type, SubsList),
	linkfeatures(Type, SubsList, HeadFeatures, TraceFeatures, trace),
                               % from module linkinfo
	add_features_to_both(HeadNr, HeadFeatures, TraceNr, TraceFeatures),
	new_nr(trace, TraceIndex), % from module linkage2links
	add_features_to_both(HeadNr, [trace=TraceIndex], TraceNr,
                         [trace=TraceIndex]),
	retract(link(HeadNr, TraceNr, Type, SubsList)),
	trace(HeadNr).

% case2: no trace exists
trace(_HeadNr).

%--------------------------------------------------+
% projection(+ProjectionType, +HeadNr, -ProjectedTree)
%--------------------------------------------------+

% case1: does a link from this head and of this projection type exist?
% yes. -> add the features defined for its link type to the feature
% structures of its head and its dependents; retract the link; project
% the head to a new node along the projection mode defined for this
% projection type; convert the dependent and all its dependents
% (recursively) to a subtree; check if this projection type is
% recursive.

projection(ProjType, HeadNr, [ProjNr, [DepProjNr|DepTree]|RestTree]):-
	link(HeadNr, DepNr, Type, SubsList),
	linkfeatures(Type, SubsList, HeadFeatures, DepFeatures, ProjType),
	add_features_to_both(HeadNr, HeadFeatures, DepNr, DepFeatures),
	retract(link(HeadNr, DepNr, Type, SubsList)),
	project(HeadNr, ProjType, ProjNr),
	merge2tree(DepNr, [DepProjNr|DepTree]),
	apply_ffp(DepProjNr, ProjNr),
	recursive_projection(ProjType, HeadNr, ProjNr, RestTree).

% case2: does a link from this head and of this projection type exist?
% no. -> Is this projection type obligatory? yes. -> project its head
% (whithout any dependents) to a new node along the projection mode
% defined for this projection type; go to the links of the next (lower)
% projection type

projection(ProjType, HeadNr, [ProjNr, NextSubTree]):-
	get_treelevel(ProjType, _Recursive, obligatory, _NextProjType,
                  _ProjFeatures),
	project(HeadNr, ProjType, ProjNr),
	next_projection(ProjType, HeadNr, NextSubTree).

% case2: does a link from this head and of this projection type exist?
% no. -> Is this projection type obligatory? no. -> go to the links of
% the next (lower) projection type.

projection(ProjType, HeadNr, NextSubTree):-
	next_projection(ProjType, HeadNr, NextSubTree).
	

%---------------------------------------------+
% project(+HeadNr, +ProjType, -ProjNr)
%---------------------------------------------+
% projects the token and creates a new avm for its projection
%- get the features which have to be added (updated) to the projection
%(eg. bar=1)
% - create a new matrix
% - add the headfeatures to the new matrix
% - add the footfeatures to the new matrix
% - add (update) the features from above to the avm of the projection
% - return the ID of the projection

project(HeadNr, ProjType, ProjNr):-
	get_treelevel(ProjType, _Recursive, _Obligatory, _NextProjType,
				  ProjFeatures),
	create_matrix([], ProjNr),
	apply_hfp(HeadNr, ProjNr),
	apply_ffp(HeadNr, ProjNr),
	add_featurelist_to_matrix(ProjNr, ProjFeatures).

%------------------------------------------------------------+
% recursive_projection(+ProjType, +HeadNr, +ProjNr, -ProjTree
%------------------------------------------------------------+

% case1: ProjType produces stacked structure if multiple links occur
% Does another link of this head and projection type exist && is this
% projection type recursive? yes and it produces stacked
% structures. -> ...
recursive_projection(ProjType, HeadNr, _ProjNr, [ProjTree]):-
	link(HeadNr, _DepNr, Type, SubsList),
	linkfeatures(Type, SubsList, _HeadFeatures, _DepFeatures, ProjType),
	get_treelevel(ProjType, stack, _Obligatory, _NextProjType,
				  _ProjFeatures),
	projection(ProjType, HeadNr, ProjTree).

% case 2: ProjType produces flat structure if multiple links occur
recursive_projection(ProjType, HeadNr, HeadProjNr, [[DepProjNr|DepTree]|RestTree]):-
	link(HeadNr, DepNr, Type, SubsList),
	linkfeatures(Type, SubsList, HeadFeatures, DepFeatures, ProjType),
	get_treelevel(ProjType, flat, _Obligatory, _NextProjType,
				  _ProjFeatures),
	add_features_to_both(HeadNr, HeadFeatures, DepNr, DepFeatures),
	retract(link(HeadNr, DepNr, Type, SubsList)),
	merge2tree(DepNr, [DepProjNr|DepTree]),
	apply_ffp(DepProjNr, HeadProjNr),
	recursive_projection(ProjType, HeadNr, HeadProjNr, RestTree).

% case3: ProjType is not recursive or no links of this type are left
recursive_projection(ProjType, HeadNr, _ProjNr, [NextSubTree]):-
	next_projection(ProjType, HeadNr,  NextSubTree).

%---------------------------------------------------+
% next_projection(+PrevProjType, +HeadNr, -ProjTree)
%---------------------------------------------------+

% case2: lowest projection type is reached
next_projection(ProjType, HeadNr, TerminalSubTree):-
	get_treelevel(ProjType, _Recursive, _Obligatory, [], _ProjFeatures),
	add_leaf(HeadNr, TerminalSubTree).

% case1: there is a next projection type left
next_projection(ProjType, HeadNr, ProjTree):-
	get_treelevel(ProjType, _Recursive, _Obligatory, NextProjType,
				  _ProjFeatures),
	projection(NextProjType, HeadNr, ProjTree).

%------------------------------------+
% add_leaf(+LeafNr, -TerminalSubTree)
%------------------------------------+

add_leaf(LeafNr, [LeafNr, LeafWord]):-
	trace(LeafNr),
	token(LeafWord, LeafNr).

%--------------------------------------------------------------------------------+
% add_features_to_both(+HeadNr, +HeadFeautures, +DependentNr, +DependentFeatures)
%--------------------------------------------------------------------------------+

add_features_to_both(HeadNr, HeadFeatures, DependentNr, DependentFeatures):-
	add_featurelist_to_matrix(HeadNr, HeadFeatures), % from module features
	add_featurelist_to_matrix(DependentNr, DependentFeatures). % from features

%-----------------------------+
% cat_is_projectable(+TokenNr)
%-----------------------------+

% if feature cat is already part of the matrix
cat_is_projectable(TokenNr):- 
	feature(TokenNr, cat=Cat),
	is_projectable(Cat).  % aus module linkinfo

cat_is_projectable(HeadNr):-   % Token ist Head
	link(HeadNr, _DepNr, Type, SubsList),
	linkfeatures(Type, SubsList, HeadFeatures, _DepFeatures,
				 _Projectiontype),
	get_feature(HeadFeatures, cat=Cat),   % from module features
	is_projectable(Cat).   % from module linkinfo

%-----------------------------------------------------+
% get_rootlink(-RootlinkType, -RootlinkSubscriptsList)
%-----------------------------------------------------+

get_rootlink(RootType, RootSubsList):-
	link(_Head, _Dependent, RootType, RootSubsList),
                                         % from module linkages2links
	is_rootlink(RootType, RootSubsList). % from module linkinfo

get_rootlink(_Rootlink, _RootSubsList):- % error message, if no root link exists
	nl, write('ERROR in linkinfo:get_rootlink/2: no rootlink found.'),
	nl, write('Execution aborted.'),
	fail.

%---------------------------------+
% apply_ffp(+DepNr, +ProjectionNr)
%---------------------------------+

apply_ffp(DepNr, ProjectionNr):- % failure-driven loop
	feature(DepNr, Feature=Value), % from module features
	is_footfeature(Feature), % from module linkinfo
	update_feature(ProjectionNr, Feature=Value),
	fail.

apply_ffp(_DepNr, _ProjectionNr).

%----------------------------------+
% apply_hfp(+HeadNr, +ProjectionNr)
%----------------------------------+

apply_hfp(HeadNr, ProjectionNr):- % failure-driven loop
	feature(HeadNr, Feature=Value), % aus module features
	is_headfeature(Feature), % aus module linkinfo
	update_feature(ProjectionNr, Feature=Value),
	fail.

apply_hfp(_HeadNr, _ProjectionNr).
