:-module(nicetree, [wordorder/2, nodelabels/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% nicetree.pl
% - rearrenges a tree according to its word order
% - labels tree
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(features, [feature/2]).
:-use_module(linkinfo, [if_option/1]).

%------------------------------------------------+
% wordorder(+RandomlyArrangedTree, -ArrangedTree)
%------------------------------------------------+

% case0: empty tree
wordorder([], []).

% case1: tree of one or more elements
wordorder([MatrixNr|DaughterTrees],
		  [MatrixNr|RecursivelySortedDaughters]):-
	quicksort(DaughterTrees, SortedDaughterTrees),
	recursivesort(SortedDaughterTrees, RecursivelySortedDaughters).

wordorder(Element, Element).

%------------------------------------------------------+
% recursivesort(+List_of_Trees, -List_of_sorted_Trees)
%------------------------------------------------------+

recursivesort([FirstTree|RestTrees],
			   [SortedFirstTree|SortedRestTrees]):-
	wordorder(FirstTree, SortedFirstTree),
	recursivesort(RestTrees, SortedRestTrees).

recursivesort([], []).

%------------------------+
% quicksort for MatrixNrs
%------------------------+

quicksort([],[]).

quicksort([X|Tail], Sorted) :-
    split(X, Tail, Small, Big),
    quicksort(Small, SortedSmall),
    quicksort(Big, SortedBig),
    append(SortedSmall, [X|SortedBig], Sorted).

split(_X, [],[],[]).
split(X, [Y|Tail], [Y|Small], Big) :-
    % X > Y, !,
	X=[XID|_], Y=[YID|_],
	feature(XID, position=PositionX),
	feature(YID, position=PositionY),
	PositionX > PositionY, !,
    split(X, Tail, Small, Big).
split(X, [Y|Tail], Small, [Y|Big]) :-
    split(X, Tail, Small, Big).

%-------------------------------------------+
% nodelabels(+UnlabelledTree, -LabelledTree)
%-------------------------------------------+

nodelabels([], []).

nodelabels([MatrixNr|Daughters], [Label|LabeledDaughters]):-
	matrixnr2label(MatrixNr, Label),
	recursivelabels(Daughters, LabeledDaughters).

nodelabels(Leaf, Leaf).

%----------------------------------------------------+
% recursivelabels(+ListofTrees, -ListofLabeledTrees)
%----------------------------------------------------+

recursivelabels([FirstTree|RestTrees],
				 [LabeledFirstTree|LabeledRestTrees]):-
	nodelabels(FirstTree, LabeledFirstTree),
	recursivelabels(RestTrees, LabeledRestTrees).

recursivelabels([], []).

%----------------------------------+
% matrixnr2label(+MatrixNr, -Label)
%----------------------------------+

matrixnr2label(MatrixNr, Label):-
	add_catbar2label(MatrixNr, CatBarLabel),
	add_trace2label(MatrixNr, CatBarLabel, TracedLabel),
	add_matrixnr2label(MatrixNr, TracedLabel, Label).

%------------------------------------+
% add_catbar2label(+MatrixNr, -Label)
%------------------------------------+

add_catbar2label(MatrixNr, Label):-
	feature(MatrixNr, cat=Cat), % wenn cat vorhanden
	name(Cat, CatList),
	feature(MatrixNr, bar=BarNr), % wenn bar vorhanden
	barnr2barletter(BarNr, BarLetter), % n2 -> np etc.
	name(BarLetter, BarList),
	append(CatList, BarList, LabelList),
	name(Label, LabelList).

add_catbar2label(MatrixNr, Label):-
	feature(MatrixNr, cat=Label). % wenn nur cat (ohne bar) vorhanden

add_catbar2label(_MatrixNr, ''). % wenn weder cat noch bar vorhanden

%------------------------------------+
% barnr2barletter(+BarNr, -BarLetter)
%------------------------------------+

barnr2barletter(BarNr, BarLetter):-
	if_option(bar=1), % from module linkinfo
	bar(BarNr, BarLetter).

barnr2barletter(BarNr, BarNr).

%----------------------+
% bar(BarNr, BarLetter)
%----------------------+

bar(0,'').
bar(1,'bar').
bar(2, 'p').
	
%-----------------------------------------------+
% add_trace2label(+Matrix, +Label, -TracedLabel)
%-----------------------------------------------+

% case1: if trace exists and trace-display-option on
add_trace2label(Matrix, Label, TracedLabel):-
	if_option(display_traces=1), % aus module linkinfo
	feature(Matrix, trace=TraceNr),
	atom_chars(Label, LabelList1),
	name(TraceNr, TraceNrList),
	atom_chars('.t', PreTraceList),
	atom_chars('', PostTraceList),
	append(LabelList1, PreTraceList, LabelList2),
	append(LabelList2, TraceNrList, LabelList3),
	append(LabelList3, PostTraceList, LabelList4),
	atom_chars(TracedLabel, LabelList4).

% case2: if no trace exists
add_trace2label(_Matrix, Label, Label).

%--------------------------------------------------------+
% add_matrixnr2label(+Matrix, +Label, -LabelInclMatrixNr)
%--------------------------------------------------------+

add_matrixnr2label(Matrix, Label, CompleteLabel):-
	if_option(display_nodeid=1), % aus module linkinfo
	atom_chars(Label, LabelList),
	name(Matrix, MatrixList),
	atom_chars('<', OpenBrList),
	atom_chars('>', CloseBrList),
	append(OpenBrList, MatrixList, LabelList1),
	append(LabelList1, CloseBrList, LabelList2),
	append(LabelList2, LabelList, LabelList4),
	atom_chars(CompleteLabel, LabelList4).

add_matrixnr2label(_Matrix, Label, Label).

%=====================+
% auxiliary predicates
%=====================+

%-------------------------------------+
% append(+List1, +List2, -CombinedList
%-------------------------------------+

append([], List, List).
append([Element], List, [Element|List]).
append([Element|Rest], List, [Element|AppendList]):-
	append(Rest, List, AppendList).

